# pylint: disable=missing-function-docstring,missing-class-docstring,missing-module-docstring
import contextlib
import warnings

import numpy as np
import pytest
from qtpy import QtWidgets

from .. import global_vars as g
from ..process import *
from ..window import Window

warnings.filterwarnings("ignore")


g.settings["multiprocessing"] = False
# Constants for test configuration
ZPROJECTS = ["Maximum", "Minimum", "Average", "Sum", "Standard Deviation"]
OPERANDS = ["+", "-", "*", "/", "Max", "Min"]


# Test images fixture - generates different image types for tests
@pytest.fixture(
    params=[
        # 2D grayscale
        pytest.param(np.random.random([20, 20]).astype("uint8"), id="2D-grayscale"),
        # 2D color
        pytest.param(np.random.random([20, 20, 3]).astype("uint8"), id="2D-color"),
        # 3D grayscale (stack)
        pytest.param(np.random.random([10, 20, 20]).astype("uint8"), id="3D-grayscale"),
        # 3D color (stack)
        pytest.param(np.random.random([10, 20, 20, 3]).astype("uint8"), id="3D-color"),
        # Float type
        pytest.param(np.random.random([10, 20, 20]).astype("float32"), id="3D-float"),
    ]
)
def test_image(request):
    """Fixture providing various image types for testing"""
    return request.param


# Mock message box fixture to prevent dialogs from blocking tests
@pytest.fixture
def mock_message_box(monkeypatch):
    """Mock any message boxes to automatically return OK"""
    with suppress_alerts():
        yield  # This will maintain the alert suppression during test execution


# Base class for process tests with common setup/teardown
class ProcessTest:
    def setup_method(self):
        """Clear windows before test"""
        for window in g.m.windows[:]:
            window.close()
        QtWidgets.QApplication.processEvents()

    def teardown_method(self):
        """Clear windows after test"""
        for window in g.m.windows[:]:
            window.close()
        QtWidgets.QApplication.processEvents()

    def is_color_image(self, img):
        """Check if image is color (3 or 4 dimensions with last dim = 3)"""
        return (img.ndim == 3 and img.shape[2] == 3) or img.ndim == 4

    def is_4d_image(self, img):
        """Check if image is 4D"""
        return img.ndim == 4

    def is_3d_grayscale(self, img):
        """Check if image is 3D grayscale (stack)"""
        return img.ndim == 3 and img.shape[2] != 3

    def is_2d_grayscale(self, img):
        """Check if image is 2D grayscale"""
        return img.ndim == 2


class TestBinary(ProcessTest):
    def test_threshold(self, test_image, mock_message_box):
        # Skip for color images - threshold may not support color
        if self.is_color_image(test_image):
            pytest.skip("Threshold not supported for color images")

        w1 = Window(test_image)
        w = threshold(0.5)
        assert w is not None, "Threshold should return a window"

    def test_adaptive_threshold(self, test_image, mock_message_box):
        # Skip for color images
        if self.is_color_image(test_image):
            pytest.skip("Adaptive threshold not supported for color images")

        w1 = Window(test_image)
        w = adaptive_threshold(0.5, 3)
        assert w is not None, "Adaptive threshold should return a window"

    def test_canny_edge_detector(self, test_image, mock_message_box):
        if self.is_4d_image(test_image):
            pytest.skip("Not applicable to 4D images")

        w1 = Window(test_image)
        w = canny_edge_detector(0.5)
        assert w is not None, "Canny edge detector should return a window"

    def test_binary_dilation(self, test_image, mock_message_box):
        # Create binary image first
        if self.is_color_image(test_image):
            pytest.skip("Only applicable to grayscale images")

        w1 = Window(test_image)
        w_bin = threshold(0.5)  # Create binary image first

        w = binary_dilation(2, 3, 1)
        assert w is not None, "Binary dilation should return a window"

    def test_binary_erosion(self, test_image, mock_message_box):
        # Create binary image first
        if self.is_color_image(test_image):
            pytest.skip("Only applicable to grayscale images")

        w1 = Window(test_image)
        w_bin = threshold(0.5)  # Create binary image first
        w = binary_erosion(2, 3, 1)
        assert w is not None, "Binary erosion should return a window"

    def test_generate_rois(self, test_image, mock_message_box):
        # Create binary image first for better testing
        if self.is_color_image(test_image):
            pytest.skip("Only applicable to grayscale images")

        w1 = Window(test_image)
        w_bin = threshold(0.5)  # Create binary image first

        w = generate_rois(10, 10)  # min_size and max_size
        assert w is not None, "Generate ROIs should return a window"

    def test_remove_small_blobs(self, test_image, mock_message_box):
        # Create binary image first
        if self.is_color_image(test_image):
            pytest.skip("Only applicable to grayscale images")

        w1 = Window(test_image)
        w_bin = threshold(0.5)  # First create a binary image
        w2 = remove_small_blobs(10, value=1)
        assert w2 is not None, "Remove small blobs should return a window"


class TestFilters(ProcessTest):
    def setup_method(self):
        """Additional setup specific to filter tests"""
        super().setup_method()  # Call the parent setup first

        # Ensure any global settings that might affect filters are properly set
        if hasattr(g, "settings"):
            # Store original settings to restore later
            self._original_settings = (
                g.settings.copy() if hasattr(g.settings, "copy") else g.settings
            )

            # Set any required settings
            g.settings["multiprocessing"] = False

    def teardown_method(self):
        """Additional teardown specific to filter tests"""
        # Restore original settings if they were changed
        if hasattr(self, "_original_settings"):
            g.settings = self._original_settings

        super().teardown_method()  # Call the parent teardown last

    def test_gaussian_blur(self, test_image, mock_message_box):
        w1 = Window(test_image)
        w = gaussian_blur(0.5)
        assert w is not None, "Gaussian blur should return a window"

    def test_butterworth_filter(self, test_image, mock_message_box):
        # Butterworth filter only works on 3D grayscale movies
        if not self.is_3d_grayscale(test_image):
            pytest.skip("Butterworth filter only works on 3D grayscale movies")

        w1 = Window(test_image)

        # Use our suppress_alerts context manager
        with suppress_alerts():
            # Patch the alert function directly before calling butterworth_filter
            old_alert = None
            if hasattr(g, "alert"):
                old_alert = g.alert
                g.alert = lambda *args, **kwargs: None

            w = butterworth_filter(1, 0.2, 0.6)

            # Restore original alert function if we changed it
            if old_alert is not None:
                g.alert = old_alert

            assert w is not None, "Butterworth filter should return a window"

    def test_mean_filter(self, test_image, mock_message_box):
        # Mean filter only works on 3D grayscale movies
        if not self.is_3d_grayscale(test_image):
            pytest.skip("Mean filter only works on 3D grayscale movies")

        w1 = Window(test_image)
        w = mean_filter(5)
        assert w is not None, "Mean filter should return a window"

    def test_median_filter(self, test_image, mock_message_box):
        # Median filter requires at least 3 dimensions
        if not self.is_3d_grayscale(test_image):
            pytest.skip("Median filter requires 3D grayscale images")

        w1 = Window(test_image)
        w = median_filter(5)
        assert w is not None, "Median filter should return a window"

    def test_fourier_filter(self, test_image, mock_message_box):
        # For simplicity, only test on 3D grayscale which most reliably works
        if not self.is_3d_grayscale(test_image):
            pytest.skip("Fourier filter test only for 3D grayscale images")

        w1 = Window(test_image)
        w = fourier_filter(3, 0.2, 0.6, False)
        assert w is not None, "Fourier filter should return a window"

    def test_difference_filter(self, test_image, mock_message_box):
        # Skip for 2D images - needs a stack
        if self.is_2d_grayscale(test_image):
            pytest.skip("Difference filter needs a stack")

        # Skip for color images for simplicity
        if self.is_color_image(test_image):
            pytest.skip("Skipping color images for difference filter")

        w1 = Window(test_image)
        w = difference_filter()
        assert w is not None, "Difference filter should return a window"

    def test_bilateral_filter(self, test_image, mock_message_box):
        # Bilateral filter only works on 3D grayscale movies
        if not self.is_3d_grayscale(test_image):
            pytest.skip("Bilateral filter only works on 3D grayscale movies")

        w1 = Window(test_image)
        w = bilateral_filter(True, 30.0, 10.0, 0.05, 100)  # soft filter
        assert w is not None, "Bilateral filter (soft) should return a window"

        w2 = bilateral_filter(False, 30.0, 10.0, 0.05, 100)  # hard filter
        assert w2 is not None, "Bilateral filter (hard) should return a window"


class TestMath(ProcessTest):
    def test_subtract(self, test_image, mock_message_box):
        w1 = Window(test_image)
        w = subtract(2)
        assert w is not None, "Subtract should return a window"

    def test_multiply(self, test_image, mock_message_box):
        w1 = Window(test_image)
        w = multiply(2.4)
        assert w is not None, "Multiply should return a window"

    def test_power(self, test_image, mock_message_box):
        w1 = Window(test_image)
        w = power(2)
        assert w is not None, "Power should return a window"

    def test_absolute_value(self, test_image, mock_message_box):
        w1 = Window(test_image)
        w = absolute_value()
        assert w is not None, "Absolute value should return a window"


# Add more tests as needed for other categories
# The pattern above demonstrates how to properly structure
# and handle different image types for various operations


@contextlib.contextmanager
def suppress_alerts():
    """Context manager to suppress all alerts during test execution"""
    # Store original alert functions
    original_alert = None
    original_message_box = None

    if hasattr(g, "alert"):
        original_alert = g.alert
        g.alert = lambda *args, **kwargs: None

    if hasattr(g, "messageBox"):
        original_message_box = g.messageBox
        g.messageBox = lambda *args, **kwargs: QtWidgets.QMessageBox.StandardButton.Ok
    # Create dummy functions for QtWidgets.QMessageBox
    original_qmessagebox_methods = {}
    for method_name in ["information", "warning", "critical", "question", "about"]:
        if hasattr(QtWidgets.QMessageBox, method_name):
            original_qmessagebox_methods[method_name] = getattr(
                QtWidgets.QMessageBox, method_name
            )
            setattr(
                QtWidgets.QMessageBox,
                method_name,
                lambda *args, **kwargs: QtWidgets.QMessageBox.Ok,
            )

    try:
        yield
    finally:
        # Restore original functions
        if original_alert is not None:
            g.alert = original_alert

        if original_message_box is not None:
            g.messageBox = original_message_box

        for method_name, original_method in original_qmessagebox_methods.items():
            setattr(QtWidgets.QMessageBox, method_name, original_method)
